#include "octree.hpp"

using namespace Quantizer;

////////////////////////////////////////
// Octree
// CTOR
Octree::Octree(const unsigned int& iColorCount, const unsigned int& iBits) :
    colorCount(iColorCount),
    leafCount (0),
    bits      (iBits)
{
    candidates = new Node* [bits+1];
    candidates[0] = root = new Internal(0, 0);
    unsigned int i;
    for(i=1; i<=bits; ++i) {
        candidates[i] = 0;
    }
}

// DTOR
Octree::~Octree() {
	Node* tmp;
    unsigned int i;
    for(i=0; i<=bits; ++i) {
        while(candidates[i]) {
            tmp = candidates[i];
            candidates[i] = candidates[i]->next;
            delete tmp;
        }
    }
    delete [] candidates;
}

// Add node to list
void Octree::Add(const unsigned int& idx, Node* node) {
	node->next = candidates[idx];
	if(candidates[idx]) {
		candidates[idx]->previous = node;
	}
	candidates[idx] = node;
}
        
// Add a color to the octree
void Octree::Add(const unsigned char& iR,
                 const unsigned char& iG, 
                 const unsigned char& iB) {
    unsigned int i, index;
    Internal     *current = dynamic_cast<Internal*>(root);
    // Traverse tree
    for(i=0; i<bits-1; ++i) {
        index = (((iR >> (7-i))<<2) & 0x04) |
                (((iG >> (7-i))<<1) & 0x02) | 
                ( (iB >> (7-i))     & 0x01);
        if(!current->child[index]) {
            current->child[index] = new Internal(index, current);
			Add(i+1, current->child[index]);
        }

        current = dynamic_cast<Internal*>(current->child[index]);
    }

	// Add leaf
	index = ((iR<<2) & 0x04) | ((iG<<1) & 0x02) | (iB & 0x01);
	Leaf* last;
	if(!current->child[index]) {
		last = new Leaf(index, current);
		Add(bits, last);
		++leafCount;
		
		current->child[index] = dynamic_cast<Node*>(last);
	} else {
		last = dynamic_cast<Leaf*>(current->child[index]);
	}
	last->Add(iR, iG, iB);
}

// Perform quantization
void Octree::Realize() {
    unsigned int i, c;
    Node *tmp, *current;
    i = bits-1;
    tmp = candidates[i];
    while((leafCount > colorCount) && (i>0)){
    	current = tmp;
    	tmp     = tmp->next;
    	
		if(current->HasChildren()) {
			Leaf* child = dynamic_cast<Internal*>(current)->Reduce(candidates + (i+1), c);
			if(child) {
				Add(bits, child);
				if(c) {
					leafCount -= c-1;
				}
				
				if(current->next) {
					current->next->previous = current->previous;
				}
				if(current->previous) {
					current->previous->next = current->next;
				}
				if(current == candidates[i]) {
					candidates[i] = current->next;
				}

				delete current;
			}
		}
		if(!tmp) {
			--i;
			tmp = candidates[i];
		}
	}
}
            
// Data accessors
unsigned int Octree::ColorCount() const {
    return colorCount;
}
unsigned int Octree::BitCount() const {
    return bits;
}

// Compute palette
unsigned char* Octree::Palette(unsigned int& paletteCount) {
    paletteCount = leafCount;
    unsigned char *palette = new unsigned char[leafCount * 3];

    unsigned int i, j;
    Node *tmp;
    Leaf *current;
    for(tmp=candidates[bits], i=0; tmp && (i<leafCount); tmp=tmp->next, ++i) {
		current = dynamic_cast<Leaf*>(tmp);
		for(j=0; j<3; ++j) {
			palette[(i*3)+j] = current->color[j] / current->reference;
		}
		current->reference = i + 1;
    }

	return palette;
}
// Get palette index 
unsigned int Octree::Index(const unsigned char& iR,
                           const unsigned char& iG,
                           const unsigned char& iB) {
    unsigned int i, index;
    unsigned char mask = 128;
    Node* tmp = root;
    i = 0;
    do {
		if(tmp->HasChildren()) {
			index = (((iR >> (7-i))<<2) & 0x04) |
					(((iG >> (7-i))<<1) & 0x02) | 
					( (iB >> (7-i))     & 0x01);
			Internal* current = dynamic_cast<Internal*>(tmp);
			tmp = current->child[index];
		} else {
			return dynamic_cast<Leaf*>(tmp)->reference - 1;
		}
		++i;
		mask >>= 1;
	} while(tmp);
    return 0;
}


////////////////////////////////////////
// Octree node
// CTOR
Octree::Node::Node(const unsigned char& idx, Node* p) :
	index   (idx),
	parent  (p),
    previous(0),
    next    (0)
{}
// DTOR
Octree::Node::~Node() {}

////////////////////////////////////////
// Leaf node
// CTOR
Octree::Leaf::Leaf(const unsigned char& idx, Node* p) :
    Node(idx, p),
    reference(0)
{
    color[0] = color[1] = color[2] = 0;
}
// DTOR
Octree::Leaf::~Leaf() {}

// Leaves don't have children
bool Octree::Leaf::HasChildren() { return false; }

// Add color to leaf
void Octree::Leaf::Add(const unsigned char& r,
					   const unsigned char& g,
	                   const unsigned char& b) {
    color[0] += r;
    color[1] += g;
    color[2] += b;
    ++reference;
}

////////////////////////////////////////
// Internal node
// CTOR
Octree::Internal::Internal(const unsigned char& idx, Node* p) :
    Node(idx, p)
{
    child[0] = child[1] = child[2] = child[3] =
    child[4] = child[5] = child[6] = child[7] = 0;
}
// DTOR
Octree::Internal::~Internal() {}

// Internal nodes at least have one child
bool Octree::Internal::HasChildren() { return true; }

// Reduce
Octree::Leaf* Octree::Internal::Reduce(Node** list, unsigned int& count) {
    Leaf *leaf = 0, *current;
    unsigned int i;
    count = 0;

    // Don't reduce "real" internal node
    for(i=0; i<8; ++i) {
    	if(!child[i]) continue;
    	if(child[i]->HasChildren())
			return 0;
    }
    
    // Let's merge children
    for(i=0; i<8; ++i) {
        if(child[i]) {
        	if(!leaf) {
        		leaf = new Leaf(index, parent);
        		if(parent) {
        			dynamic_cast<Internal*>(parent)->child[index] = leaf;
        		}
        	}
			current = dynamic_cast<Leaf*>(child[i]);
			leaf->reference += current->reference;
			leaf->color[0]  += current->color[0];
			leaf->color[1]  += current->color[1];
			leaf->color[2]  += current->color[2];

			if(child[i]->previous) {
				child[i]->previous->next = child[i]->next;
			}

			if(*list == child[i]) {
				*list = child[i]->next;
			}

			if(child[i]->next) {
				child[i]->next->previous = child[i]->previous;
			}

			++ count;

			delete child[i];
			child[i] = 0;
        }
    }
    
    return leaf;
}

